#!/usr/bin/env python
from datatypes import *
from fp_tools import *
import sys
import struct
import math
import numpy as np
from PIL import Image
from rt_core import rt_core


def get_initial_params(ray_orig, ray_dir):
    ray_len = (ray_dir[0]**2 + ray_dir[1]**2 + ray_dir[2]**2)**0.5
    ray_dir[0] /= ray_len
    ray_dir[1] /= ray_len
    ray_dir[2] /= ray_len

    a = 0
    dim = 1

    for comp in range(0, 3):
        if ray_dir[comp] == 0:
            ray_dir[comp] = 0.0001
        if ray_dir[comp] < 0:
            ray_orig[comp] *= -1
            ray_dir[comp] *= -1
            a |= 2**comp

    t0_float = (
        (-dim - ray_orig[0]) / ray_dir[0],
        (-dim - ray_orig[1]) / ray_dir[1],
        (-dim - ray_orig[2]) / ray_dir[2]
    )

    t1_float = (
        (dim - ray_orig[0]) / ray_dir[0],
        (dim - ray_orig[1]) / ray_dir[1],
        (dim - ray_orig[2]) / ray_dir[2]
    )

    t0 = vec3(t0_float[0], t0_float[1], t0_float[2])
    t1 = vec3(t1_float[0], t1_float[1], t1_float[2])

    return {'a': a, 't0': t0, 't0_float': t0_float, 't1': t1, 't1_float': t1_float}


reset = False
start = False

out = {}
max_it = 0
max_params = {}

def run_rt_core(ray_orig, ray_dir, debug=False):
    global out
    global max_it
    global max_params
    params = get_initial_params(ray_orig.copy(), ray_dir.copy())
    params['ray_orig'] = ray_orig
    params['ray_dir'] = ray_dir

    if fp.gte(params['t0'].max(), params['t1'].min()):
        out['cost'] = 0
        return False

    params['entry'] = 0

    out = rt_core(True, False, {}, mem, debug)
    out = rt_core(False, True, params, mem, debug)

    for i in range(0, 1000):
        out = rt_core(False, False, params, mem, debug)
        if out['ready']:
            hit = out['hit']
            out = rt_core(False, False, {}, mem, debug)
            out = rt_core(False, False, {}, mem, debug)
            if hit and i > max_it:
                max_it = i
                max_params = params
            out['cost'] = i
            return hit



def get_inv_proj(width, height):
    far = 100.0
    near = 0.001
    fov = 75.0
    aspect = float(width) / height

    f = far
    n = near
    t = math.tan(fov / 180 * math.pi / 2) * n
    b = -t
    r = t * aspect
    l = -r


    return np.array([[-f * (r - l) / (f - n), 0, 0, -f * (r + l) / (f - n)],
            [0, -f * (t - b) / (f - n), 0, -f * (t + b) / (f - n)],
            [0, 0, 0, 2 * f * n / (f - n)],
            [0, 0, 1, -(f + n) / (f - n)]])


def get_inv_view(pos, rot):
    inv_view = np.array([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1]
    ])

    rotation_y = np.array([
        [math.cos(-rot[1]), 0, math.sin(-rot[1]), 0],
        [0, 1, 0, 0],
        [-math.sin(-rot[1]), 0, math.cos(-rot[1]), 0],
        [0, 0, 0, 1]
    ])

    rotation_x = np.array([
        [1, 0, 0, 0],
        [0, math.cos(0.5* math.sin(rot[0])), -math.sin(0.5* math.sin(rot[0])), 0],
        [0, math.sin(0.5* math.sin(rot[0])), math.cos(0.5* math.sin(rot[0])), 0],
        [0, 0, 0, 1]
    ])

    rotation = rotation_x.dot(rotation_y)

    translation = np.array([[1, 0, 0, pos[0]],
                   [0, 1, 0, pos[1]],
                   [0, 0, 1, pos[2]],
                   [0, 0, 0, 1]])

    inv_view = np.transpose(rotation).dot(inv_view)
    inv_view = translation.dot(inv_view)
    return inv_view


def get_ray(x, y, width, height, pos, rot):
    v0 = np.array([x, y, -1, 1])
    v1 = np.array([x, y, 1, 1])

    inv_proj = get_inv_proj(width, height)
    inv_view = get_inv_view(pos, rot)

    q0 = inv_proj.dot(v0)
    q1 = inv_proj.dot(v1)

    q0 = q0 / q0[3]
    q1 = q1 / q1[3]

    q0 = inv_view.dot(q0)
    q1 = inv_view.dot(q1)

    ray_orig = [q0[0], q0[1], q0[2]]
    ray_dir = [q1[0] - q0[0], q1[1] - q0[1], q1[2] - q0[2]]

    return {'orig': ray_orig, 'dir': ray_dir}

def get_ray_opt(x, y, theta):
    P = get_inv_proj()
    a_coef = -2 / (P[3, 3]**2 - 1)
    a_x = a_coef * P[0, 0]
    a_y = a_coef * P[1, 1]
    a_z = a_coef * P[2, 3]
    b_x = a_coef * P[0, 3]
    b_y = a_coef * P[1, 3]

    c_coef = 1 / (P[3, 3] - 1)
    c_x = c_coef * P[0, 0]
    c_y = c_coef * P[1, 1]
    c_z = c_coef * P[2, 3]
    d_x = c_coef * P[0, 3]
    d_y = c_coef * P[1, 3]

    t_x = 0
    t_y = 0
    t_z = 2.5
    #t_x = 0
    #t_y = -0.45
    #t_z = 0.9

    ray_orig = [c_x * x + d_x + t_x, c_y * y + d_y + t_y, c_z + t_z]
    ray_dir = [a_x * x + b_x, a_y * y + b_y, a_z]

    return {'orig': ray_orig, 'dir': ray_dir}


def get_single_ray(x, y, width, height):
    ray = get_ray_opt((x + 0.5) / width * 2 - 1.0, ((height - y - 1) + 0.5) / height * 2 - 1.0, 0)
    string = '%08x %08x %08x %08x %08x %08x' % (
        get_float(ray['orig'][0]),
        get_float(ray['orig'][1]),
        get_float(ray['orig'][2]),
        get_float(ray['dir'][0]),
        get_float(ray['dir'][1]),
        get_float(ray['dir'][2])
    )
    params = get_initial_params(ray['orig'].copy(), ray['dir'].copy())

    #return string + ' %08x %08x %08x %08x %08x %08x %x\n' % (
    return string + ' %x\n' % (
    #    get_float(params['t0_float'][0]),
    #    get_float(params['t0_float'][1]),
    #    get_float(params['t0_float'][2]),
    #    get_float(params['t1_float'][0]),
    #    get_float(params['t1_float'][1]),
    #    get_float(params['t1_float'][2]),
        params['a']
    )



def dragon_mem():
    with open('../models/svo/dragon_256.svo', 'rb') as file:
        data = file.read()

    return list(struct.unpack('%di' % (len(data) / 4), data))

def basic_mem():
    return [
        0x00015A10,
        0x00041410,
        0x80021911,
        0x00042C00,
        0x00000004,
        0x00001212,
        0x00004C4C,
        0x00000606,
        0x00006161,
        0x00001010
    ]

def get_ray_tb():
    file = open('out.hex', 'w+')
    width = 256
    height = 256
    for y in range(0, height):
        for x in range(0, width):
            file.write(get_single_ray(x, y, width, height));

    file.close()




if __name__ == "__main__":
    mem = dragon_mem()
    #mem = basic_mem()

    fp.size = 32
    fp.global_prec = 16
    width = 320
    height = 180

    data = np.zeros((height, width, 3))

    string = ""

    file = open('out.hex', 'w+')

    for y in range(0, height):
        for x in range(0, width):
            ray = get_ray(
                (x + 0.5) / width * 2 - 1.0,
                ((height - y - 1) + 0.5) / height * 2 - 1.0,
                width,
                height,
                [-0.5, 0.35, -2.1],
                [math.pi * 4/12, math.pi * 13/12, 0]
            )

            params = get_initial_params(ray['orig'].copy(), ray['dir'].copy())
            t_hit = 0xFFFFFFFF
            val = 0.0
            hit = run_rt_core(ray['orig'], ray['dir'], x + y * width == 0x722b)

            if not hit:
                #data[y, x] = [1.0, 0.95**(out['cost'] / 2), 0.95**(out['cost'] / 2)]
                data[y, x] = [1.0, 1.0, 1.0]
            else:
                t_hit = out['t_hit'].binval
                if out['normal'].x.binval:
                    data[y, x] = [0.7, 0.7, 0.7]
                elif out['normal'].y.binval:
                    data[y, x] = [0.9, 0.9, 0.9]
                else:
                    data[y, x] = [0.5, 0.5, 0.5]

            file.write('%08x %08x %08x %08x %08x %08x %08x %08x\n' % (
                (x + y * width) | params['a'] << 28,
                get_float(params['t0_float'][0]),
                get_float(params['t0_float'][1]),
                get_float(params['t0_float'][2]),
                get_float(params['t1_float'][0]),
                get_float(params['t1_float'][1]),
                get_float(params['t1_float'][2]),
                #get_float(ray['orig'][0]),
                #get_float(ray['orig'][1]),
                #get_float(ray['orig'][2]),
                #get_float(ray['dir'][0]),
                #get_float(ray['dir'][1]),
                #get_float(ray['dir'][2]),
                t_hit
            ))

        if (y + 1) % 10 == 0:
            print('\tRow %d of %d' % (y + 1, height))

    img = Image.fromarray((data * 255).astype('uint8'))
    img.save('img_test.png')
    file.close()

