import types::*;

typedef enum logic[2:0] {IDLE, INIT, EVAL, NEXT, PUSH, POP, OUT} rt_core_state_t;

typedef struct packed {
    logic [31:0] entry_idx;
    vec3_t t0;
    vec3_t tm;
    vec3_t t1;
    logic [2:0] cur_child_idx;
} stack_frame_t;

module svo_trav_core (
    input  logic        clk,
    input  logic        reset,
    output logic [31:0] mem_addr,
    input  logic [31:0] mem_data,

    output logic        in_ready,
    input  logic        in_valid,
    input  job_t        in_job,

    input  logic        out_ready,
    output logic        out_valid,
    output result_t     out_result

);

rt_core_state_t core_state;
assign in_ready = (core_state == IDLE);

stack_frame_t cur_stack_frame;
stack_frame_t stack_top;
logic stack_push;
logic stack_pop;
logic stack_empty;
logic push_indirect;
logic stack_clear;
stack inst_stack(
    clk,
    reset,
    stack_clear,
    stack_push,
    stack_pop,
    cur_stack_frame,
    stack_top,
    stack_empty
);

logic [2:0] transform;
logic [2:0] cur_child_idx_t;
always_comb begin
    cur_child_idx_t = cur_stack_frame.cur_child_idx ^ transform;
end

logic [95:0] top_t0;
logic [95:0] top_t1;
logic ray_valid;
get_ray_valid inst_get_ray_valid(clk, top_t0, top_t1, ray_valid);

vec3_t t0_child;
vec3_t t1_child;
get_child_t_values inst_t0_get_child_t_values(
    cur_stack_frame.cur_child_idx,
    cur_stack_frame.t0,
    cur_stack_frame.tm,
    t0_child
);

get_child_t_values inst_t1_get_child_t_values(
    cur_stack_frame.cur_child_idx,
    cur_stack_frame.tm,
    cur_stack_frame.t1,
    t1_child
);

logic [2:0] initial_child_idx;
get_initial_child_idx inst_get_initial_child_idx(
    cur_stack_frame.t0,
    cur_stack_frame.tm,
    initial_child_idx
);

logic [3:0] next_child_idx;
get_next_child_idx inst_get_next_child_idx(
    clk,
    cur_stack_frame.cur_child_idx,
    t1_child,
    next_child_idx
);


logic [2:0] ones_count [8];
get_ones_count inst_get_ones_count(mem_data[14:8] & ~mem_data[6:0], ones_count);

logic[31:0] next_entry_idx;

assign out_valid = (core_state == OUT);

always @(posedge clk) begin
    if (reset == 1'b1) begin
        core_state      <= IDLE;
        cur_stack_frame <= 'b0;
        stack_push      <= 'b0;
        stack_pop       <= 'b0;
        stack_clear     <= 'b0;
        push_indirect   <= 'b0;
        next_entry_idx  <= 'b0;
        mem_addr        <= 'bX;
        transform       <= 'bX;
        out_result      <= 'bX;
    end else begin
        if (core_state == IDLE) begin
            out_result.cost <= 'b0;
        end else if (core_state != OUT) begin
            out_result.cost <= out_result.cost + 1;
        end

        case (core_state)
            IDLE: begin
                if (in_ready & in_valid) begin
                    next_entry_idx    <= 'b0;
                    mem_addr          <= 'b0;
                    stack_clear       <= 'b1;

                    out_result.job_id <= in_job.job_id;

                    cur_stack_frame.entry_idx <= in_job.addr;
                    cur_stack_frame.t0.x <= in_job.t0[0 +:32];
                    cur_stack_frame.t0.y <= in_job.t0[32+:32];
                    cur_stack_frame.t0.z <= in_job.t0[64+:32];
                    cur_stack_frame.t1.x <= in_job.t1[0 +:32];
                    cur_stack_frame.t1.y <= in_job.t1[32+:32];
                    cur_stack_frame.t1.z <= in_job.t1[64+:32];
                    cur_stack_frame.tm.x <= (
                        $signed(in_job.t0[0 +:32] + in_job.t1[0 +:32]) >>> 1
                    );
                    cur_stack_frame.tm.y <= (
                        $signed(in_job.t0[32+:32] + in_job.t1[32+:32]) >>> 1
                    );
                    cur_stack_frame.tm.z <= (
                        $signed(in_job.t0[64+:32] + in_job.t1[64+:32]) >>> 1
                    );

                    top_t0     <= in_job.t0;
                    top_t1     <= in_job.t1;
                    transform  <= in_job.a;
                    core_state <= INIT;
                end
            end

            INIT: begin
                cur_stack_frame.cur_child_idx <= initial_child_idx;

                mem_addr    <= cur_stack_frame.entry_idx;
                stack_clear <= 1'b0;
                core_state  <= EVAL;
            end

            EVAL: begin
                if (!ray_valid) begin
                    out_result.hit <= 1'b0;
                    core_state     <= OUT;
                end else if (mem_data[{2'b01, cur_child_idx_t}]) begin
                    if (mem_data[{2'b00, cur_child_idx_t}]) begin
                        out_result.hit    <= 1'b1;
                        out_result.normal <= 'b0;

                        if (
                            $signed(t0_child.x) >= $signed(t0_child.y) &&
                            $signed(t0_child.x) >= $signed(t0_child.z)
                        ) begin
                            out_result.t_hit          <= t0_child.x;
                            out_result.normal[0 +:32] <= 32'h00010000;
                        end else if (
                            $signed(t0_child.y) >= $signed(t0_child.x) &&
                            $signed(t0_child.y) >= $signed(t0_child.z)
                        ) begin
                            out_result.t_hit          <= t0_child.y;
                            out_result.normal[32+:32] <= 32'h00010000;
                        end else begin
                            out_result.t_hit          <= t0_child.z;
                            out_result.normal[64+:32] <= 32'h00010000;
                        end

                        core_state <= OUT;
                    end else begin
                        stack_push <= 1'b1;
                        if (mem_data[31] == 1'b1) begin
                            mem_addr      <= (
                                cur_stack_frame.entry_idx +
                                {17'b0, mem_data[30:16]}
                            );
                            push_indirect <= 1'b1;
                        end else begin
                            next_entry_idx <= (
                                cur_stack_frame.entry_idx +
                                {17'b0, mem_data[30:16]}
                            );
                        end
                        core_state <= PUSH;
                    end
                end else begin
                    core_state <= NEXT;
                end
            end

            NEXT: begin
                if (next_child_idx < 8) begin
                    cur_stack_frame.cur_child_idx <= next_child_idx[2:0];
                    core_state <= EVAL;
                end else begin
                    if (stack_empty == 1'b1) begin
                        out_result.hit <= 1'b0;
                        core_state     <= OUT;
                    end else begin
                        cur_stack_frame <= stack_top;
                        stack_pop       <= 1'b1;
                        core_state      <= POP;
                    end
                end
            end

            PUSH: begin
                stack_push      <= 1'b0;
                cur_stack_frame <= '0;
                if (push_indirect == 1'b1) begin
                    cur_stack_frame.entry_idx <= (
                        cur_stack_frame.entry_idx +
                        mem_data + {29'b0, ones_count[cur_child_idx_t]}
                    );
                end else begin
                    cur_stack_frame.entry_idx <= (
                        next_entry_idx +
                        {29'b0, ones_count[cur_child_idx_t]}
                    );
                end

                push_indirect        <= 1'b0;
                cur_stack_frame.t0   <= t0_child;
                cur_stack_frame.t1   <= t1_child;
                cur_stack_frame.tm.x <= $signed(t0_child.x + t1_child.x) >>> 1;
                cur_stack_frame.tm.y <= $signed(t0_child.y + t1_child.y) >>> 1;
                cur_stack_frame.tm.z <= $signed(t0_child.z + t1_child.z) >>> 1;
                core_state           <= INIT;
            end

            POP: begin
                stack_pop  <= 1'b0;
                mem_addr   <= cur_stack_frame.entry_idx;
                core_state <= NEXT;
            end

            OUT: begin
                if (out_ready) begin
                    core_state <= IDLE;
                end
            end
        endcase
    end
end
endmodule


module get_initial_child_idx(
    input  vec3_t      t0,
    input  vec3_t      tm,
    output logic [2:0] initial_child_idx
);

always_comb begin
    initial_child_idx = 3'h0;
    if ($signed(t0.x) >= $signed(t0.y) &&
        $signed(t0.x) >= $signed(t0.z)) begin
        if ($signed(t0.x) >= $signed(tm.y)) begin
            initial_child_idx[1] = 1'b1;
        end
        if ($signed(t0.x) >= $signed(tm.z)) begin
            initial_child_idx[2] = 1'b1;
        end
    end else if ($signed(t0.y) >= $signed(t0.x) &&
        $signed(t0.y) >= $signed(t0.z)) begin
        if ($signed(t0.y) >= $signed(tm.x)) begin
            initial_child_idx[0] = 1'b1;
        end
        if ($signed(t0.y) >= $signed(tm.z)) begin
            initial_child_idx[2] = 1'b1;
        end
    end else begin
        if ($signed(t0.z) >= $signed(tm.x)) begin
            initial_child_idx[0] = 1'b1;
        end
        if ($signed(t0.z) >= $signed(tm.y)) begin
            initial_child_idx[1] = 1'b1;
        end
    end
end
endmodule


module get_next_child_idx(
    input  logic        clk,
    input  logic [2:0]  cur_child_idx,
    input  vec3_t       t1,
    output logic [3:0]  next_child_idx
);

logic [3:0] lookup [0:31] = '{
    4'b0001, 4'b0010, 4'b0100, 4'b1111,
    4'b1000, 4'b0011, 4'b0101, 4'b1111,
    4'b0011, 4'b1000, 4'b0110, 4'b1111,
    4'b1000, 4'b1000, 4'b0111, 4'b1111,
    4'b0101, 4'b0110, 4'b1000, 4'b1111,
    4'b1000, 4'b0111, 4'b1000, 4'b1111,
    4'b0111, 4'b1000, 4'b1000, 4'b1111,
    4'b1000, 4'b1000, 4'b1000, 4'b1111
};

always @(posedge clk) begin
    if (
        $signed(t1.x) <= $signed(t1.y) &&
        $signed(t1.x) <= $signed(t1.z)
    ) begin
        next_child_idx <= lookup[{cur_child_idx, 2'b00}];
    end else if (
        $signed(t1.y) <= $signed(t1.x) &&
        $signed(t1.y) <= $signed(t1.z)
    ) begin
        next_child_idx <= lookup[{cur_child_idx, 2'b01}];
    end else begin
        next_child_idx <= lookup[{cur_child_idx, 2'b10}];
    end
end
endmodule


module get_ray_valid(
    input  logic        clk,
    input  logic [95:0] t0,
    input  logic [95:0] t1,
    output logic        ray_valid
);

logic [31:0] t0_max;
logic [31:0] t1_min;

always_comb begin
    if (
        $signed(t0[0 +:32]) >= $signed(t0[32+:32]) &&
        $signed(t0[0 +:32]) >= $signed(t0[64+:32])
    ) begin
        t0_max = t0[0 +:32];
    end else if (
        $signed(t0[32+:32]) >= $signed(t0[0 +:32]) &&
        $signed(t0[32+:32]) >= $signed(t0[64+:32])
    ) begin
        t0_max = t0[32+:32];
    end else begin
        t0_max = t0[64+:32];
    end
end

always_comb begin
    if (
        $signed(t1[0 +:32]) <= $signed(t1[32+:32]) &&
        $signed(t1[0 +:32]) <= $signed(t1[64+:32])
    ) begin
        t1_min = t1[0 +:32];
    end else if (
        $signed(t1[32+:32]) <= $signed(t1[0 +:32]) &&
        $signed(t1[32+:32]) <= $signed(t1[64+:32])
    ) begin
        t1_min = t1[32+:32];
    end else begin
        t1_min = t1[64+:32];
    end
end

always @(posedge clk) begin
    if ($signed(t0_max) >= $signed(t1_min)) begin
        ray_valid = 0;
    end else begin
        ray_valid = 1;
    end
end

endmodule


module get_ones_count(
    input  logic [6:0] data,
    output logic [2:0] ones_count [8]
);

always_comb begin
    for (int idx = 0; idx < 8; idx += 1) begin
        ones_count[idx] = '0;
        for (int sub_idx = 0; sub_idx < idx; sub_idx += 1) begin
            ones_count[idx] += {2'b00, data[sub_idx]};
        end
    end
end
endmodule


module stack(
    input   logic clk,
    input   logic reset,
    input   logic clear,
    input   logic push,
    input   logic pop,
    input   stack_frame_t data,
    output  stack_frame_t top,
    output  logic empty
);

stack_frame_t [10:0] stack;
logic [7:0] stack_idx;
always_comb begin
    empty = (stack_idx == 8'h0);
    if (empty) begin
        top   = 'b0;
    end else begin
        top   = stack[stack_idx - 1];
    end
end

always @(posedge clk) begin
    if (reset == 1'b1 || clear == 1'b1) begin
        stack_idx <= 0;
    end else if (push == 1'b1) begin
        stack[stack_idx] <= data;
        stack_idx <= stack_idx + 1;
    end else if (pop == 1'b1) begin
        stack_idx <= stack_idx - 1;
    end
end

endmodule

module get_child_t_values(
    input  logic [2:0] cur_child_idx,
    input  vec3_t      t0,
    input  vec3_t      t1,
    output vec3_t      t_child
);

always_comb begin
    if (cur_child_idx[0] == 1'b1) begin
        t_child.x = t1.x;
    end else begin
        t_child.x = t0.x;
    end
    if (cur_child_idx[1] == 1'b1) begin
        t_child.y = t1.y;
    end else begin
        t_child.y = t0.y;
    end
    if (cur_child_idx[2] == 1'b1) begin
        t_child.z = t1.z;
    end else begin
        t_child.z = t0.z;
    end
end
endmodule
